--[[
Main class for the add-on.  Handles all run-time functionality.
--]]

uab.classes.main = uab.mixins.aceOO.Class(uab.mixins.aceEvent)

function uab.classes.main.TimerEvent(self, event, elapsed)
    self[event](self, elapsed)
end

function uab.classes.main.prototype:init()
	uab.classes.main.super.prototype.init(self)

    self.name = "uab.main"	
	self.layoutManagers = {}
	self.layouts = {}
	self.activators = {}
	self.combatRegistrations = {}
	self.combatUnregistrations = {}
    self.timersEnabledCount = 0

    uab:RegisterMetro("ActionButton", uab.classes.main.TimerEvent, 0, self, "ActionButtonTimerEvent")
    uab:RegisterMetro("FrameRegister", uab.classes.main.TimerEvent, 0, self, "DelayedFrameRegisterTimerEvent")
end

function uab.classes.main.prototype:Enable()
	self:RegisterEvent("Banzai_UnitGainedAggro")
	self:RegisterEvent("Banzai_UnitLostAggro")
    self:RegisterEvent("UNIT_HEALTH", "UpdateHealth")
    
    self:ConnectToUnitFrames()

	self:EnableActivators()

	-- Make sure that we show the correct layout and at the correct scale
    self:LayoutChanged()
    self:ScaleChanged()
    
    self:GetCfgDlg():GetActionsTab():ActivatorsChanged()
end

function uab.classes.main.prototype:GetCfgDlg()
    if not self.cfgdlg then
        self.cfgdlg = uab.classes.cfgdlg:new()
    end
    
    return self.cfgdlg
end

function uab.classes.main.prototype:UpdateHealth()
	if UnitIsDead("player") then
		self:CloseAllActionSets()
	end
end

function uab.classes.main.prototype:CloseAllActionSets()
	for _, activator in pairs(self.activators) do
	    activator:IAmDead()
	end
end

function uab.classes.main.prototype:ActivatorExists(name)
    if self.activators[name] then
        return true
    end
end

function uab.classes.main.prototype:AnyActivators(name)
    local any
    
    for name, activator in pairs(self.activators) do
        if name and activator then
            any = true
            break
        end
    end
    
    return any
end

function uab.classes.main.prototype:NewActivator(name)
    local activation = self:FindNextAvailableActivation()

    if not activation then
        uab.error("No available activation click combinations available.")
        return
    end
        
    local newactivator = {}

	newactivator.enabled = true
    newactivator.button = activation.button
    newactivator.prefixshift = activation.prefixshift
    newactivator.prefixctrl = activation.prefixctrl
    newactivator.prefixalt = activation.prefixalt
    
    newactivator.groups = {}
    for _, info in pairs(uab.constants.groupInfo) do
        newactivator.groups[info.name] = {}
        newactivator.groups[info.name].enabled = true
    end
    
    uab.db.profile.activators[name] = newactivator
    
	self.activators[name] = uab.classes.ClickActivator:new(name, uab.db.profile.activators[name])
    self.activators[name]:Enable()

	--[[
	Hook this activator to every frame that is registered
	--]]
    for frame in pairs(self.ccframes) do
        self.activators[name]:RegisterFrame(frame)
    end

    uab:RefreshActivatorList()
    
    self:GetCfgDlg():GetActionsTab():ActivatorsChanged()
end

function uab.classes.main.prototype:GetActivator(name)
    return self.activators[name]
end

function uab.classes.main.prototype:CopyActivatorTo(from, to)
    self.activators[from]:CopyTo(self.activators[to])
end

function uab.classes.main.prototype:RemoveActivator(name)
	self.activators[name]:Disable()
	self.activators[name] = nil

	uab.db.profile.activators[name] = nil
	
	uab:RefreshActivatorList()
    
    self:GetCfgDlg():GetActionsTab():ActivatorsChanged()
end

function uab.classes.main.prototype:AddActivatorKey(name, key)
	local states = self:DisableActivators()
	
    uab.db.profile.activators[name]["prefix"..key] = true
    self.activators[name]:SetActivationData()

	self:EnableActivators(states)
end

function uab.classes.main.prototype:RemoveActivatorKey(name, key)
	local states = self:DisableActivators()

    uab.db.profile.activators[name]["prefix"..key] = nil
    self.activators[name]:SetActivationData()

	self:EnableActivators(states)
end

function uab.classes.main.prototype:ClearActivatorKeys(name)
	local states = self:DisableActivators()

    uab.db.profile.activators[name].prefixalt = nil
    uab.db.profile.activators[name].prefixctrl = nil
    uab.db.profile.activators[name].prefixshift = nil
    self.activators[name]:SetActivationData()

	self:EnableActivators(states)
end

function uab.classes.main.prototype:SetActivatorButton(name, btn)
	local states = self:DisableActivators()

    uab.db.profile.activators[name].button = btn
    self.activators[name]:SetActivationData()

	self:EnableActivators(states)
end

function uab.classes.main.prototype:DisableActivators()
	local states = {}
	for _, activator in pairs(self.activators) do
	    states[activator] = activator:Enabled()
		activator:Disable()
	end
	
	return states
end

function uab.classes.main.prototype:EnableActivators(states)
	for _, activator in pairs(self.activators) do
	    if not states or states[activator] then
			activator:Enable()
		end
	end
end

function uab.classes.main.prototype:DisableActivator(name)
    if self.activators[name] and self.activators[name]:Enabled() then
        self.activators[name]:Disable()
    end
end

function uab.classes.main.prototype:EnableActivator(name)
    if self.activators[name] and not self.activators[name]:Enabled() then
        self.activators[name]:Enable()
    end
end

function uab.classes.main.prototype:DisableActionButtonTimer()
    self.timersEnabledCount = self.timersEnabledCount - 1
    if self.timersEnabledCount < 0 then
        self.timersEnabledCount = 0
    end
    
    if self.timersEnabledCount == 0 then
        uab:StopMetro("ActionButton")
    end
end

function uab.classes.main.prototype:EnableActionButtonTimer()
    self.timersEnabledCount = self.timersEnabledCount + 1

    if self.timersEnabledCount == 1 then
        uab:StartMetro("ActionButton")
    end
end

function uab.classes.main.prototype:ActionButtonTimerEvent(elapsed)
	for _, activator in pairs(self.activators) do
	    if activator:Enabled() then
            activator:TimerEvent(elapsed)
		end
	end
end

function uab.classes.main.prototype:CreateActivator(name, data)
	self.activators[name] = uab.classes.ClickActivator:new(name, data)
end

function uab.classes.main.prototype:Banzai_UnitGainedAggro(unitId)
end

function uab.classes.main.prototype:Banzai_UnitLostAggro(unitId)
end

function uab.classes.main.prototype:ConnectToUnitFrames()

    self:HookClickCastFrames()

    -- Register all frames that snuck in before we did =)
    for frame in pairs(self.ccframes) do
		self:RegisterFrame(frame)
    end
end

function uab.classes.main.prototype:CCFNewIndex(t, k, v)
	if self.ccframesHooked and not self.ccframesNewIndex then
	    if v == nil then
	        self:UnregisterFrame(k)
	    else
	        self:RegisterFrame(k)
	    end
	end
	
    --[[
	If there was someone loaded before us, and they catch
    ClickCastFrames changes, then let them have the event
    and let them do the rawset.  First, they need to process
    the event, and second, if they process it, they will
    either do the rawset or pass it to the next one in line
    as we are doing here.
    --]]
    
    if self.ccframesNewIndex then
        self.ccframesNewIndex(t,k,v)
    else
        if v == nil then
            rawset(self.ccframes, k, nil)
        else
            rawset(self.ccframes, k, v)
        end
    end
end

function uab.classes.main.prototype:HookClickCastFrames()
	if not self.ccframesHooked then
	    self.ccframesHooked = true
	    
		ClickCastFrames = ClickCastFrames or {}
		self.ccframes = ClickCastFrames

	    local mt = getmetatable(ClickCastFrames)
	    if not mt then
	        mt = {}
	    end

	    self.ccframesNewIndex = mt.__newindex

        if self.ccframesNewIndex then
            --[[
            Another click-cast add-on is already loaded.  We
            should note this, and not attach ourselves.  Any
            further actions will be retained in case at some
            reload, the other click-cast add-on is not loaded
            but the actions will have no effect in the current
            session.
            --]]
            
            uab.print("Another click-casting add-on is already loaded.")
            uab.print("If you want to use UnitActionBars, please disable the other add-on.")
            uab.print("You may set up action sets, actions, and general settings, but those")
            uab.print("settings will not have any effect, as UnitActionBars will not attach")
            uab.print("to any frame to enable click-casting.")
        end
        
	    local newindex = function(t,k,v)
	        uab.main:CCFNewIndex(t, k, v)
	    end

	    mt.__newindex = newindex

	    ClickCastFrames = setmetatable(ClickCastFrames, mt)
	end
	
    local tbl = {
		PlayerFrame,
		PetFrame,
		PartyMemberFrame1,
		PartyMemberFrame2,
		PartyMemberFrame3,
		PartyMemberFrame4,
		PartyMemberFrame1PetFrame,
		PartyMemberFrame2PetFrame,
		PartyMemberFrame3PetFrame,
		PartyMemberFrame4PetFrame,
		TargetFrame,
		TargetofTargetFrame,
    }

    for i,frame in pairs(tbl) do
		if not ClickCastFrames[frame] then
            ClickCastFrames[frame] = true
        end
    end
end

function uab.classes.main.prototype:Disable()
	self:DisableActivators()
    self:DisconnectFromUnitFrames()

	self:UnregisterEvent("Banzai_UnitGainedAggro")
	self:UnregisterEvent("Banzai_UnitLostAggro")
end

function uab.classes.main.prototype:DisconnectFromUnitFrames()

	-- stop us from registering anyone while disconnecting
    self:UnhookClickCastFrames()

    -- Unregister all frames that are currently registered
    for frame in pairs(self.ccframes) do
		self:UnregisterFrame(frame)
    end
end

function uab.classes.main.prototype:UnhookClickCastFrames()
	--[[
	We don't actually disconnect from the metatable, as someone may
	have connected after us.  Instead, we just indicate that we should
	not process any registrations now.  They will still be in the table
	when we are re-enabled, and will be registered at that time.
	--]]

	self.ccframesHooked = nil
end

function uab.classes.main.prototype:RegisterFrame(frame)
    if InCombatLockdown() then
        self:DelayRegisterFrame(frame)
        return
    end

    uab.pdebug("Registering frame : "..frame:GetName())
    for _, activator in pairs(self.activators) do
        activator:RegisterFrame(frame)
    end
end

function uab.classes.main.prototype:UnregisterFrame(frame)
    if InCombatLockdown() then
        self:DelayUnregisterFrame(frame)
        return
    end

    for _, activator in pairs(self.activators) do
        activator:UnregisterFrame(frame)
    end
end

function uab.classes.main.prototype:DelayRegisterFrame(frame)
	self.combatRegistrations[frame] = true
    uab:StartMetro("FrameRegister")
end

function uab.classes.main.prototype:DelayUnregisterFrame(frame)
	self.combatUnregistrations[frame] = true
    uab:StartMetro("FrameRegister")
end

function uab.classes.main.prototype:DelayedFrameRegisterTimerEvent(elapsed)
    if InCombatLockdown() then
		return
	end

    self:RegisterDelayedFrames()
    self:UnregisterDelayedFrames()
    uab:StopMetro("FrameRegister")
end

function uab.classes.main.prototype:RegisterDelayedFrames()
	for frame, _ in pairs(self.combatRegistrations) do
		self:RegisterFrame(frame)
	end
end

function uab.classes.main.prototype:UnregisterDelayedFrames()
	for frame, _ in pairs(self.combatUnregistrations) do
		self:UnregisterFrame(frame)
	end
end

function uab.classes.main.prototype:LayoutChanged()
    if not uab.db.profile.layoutSettings then
        uab.db.profile.layoutSettings = {}
    end
    
    local managers = uab:GetManagers()

    local layoutname = uab.db.profile.layout
    local manager = managers[layoutname].manager
    
    if not uab.db.profile.layoutSettings[layoutname] then
        uab.db.profile.layoutSettings[layoutname] = {}
	end
	
    if not managers[layoutname].layout then
        managers[layoutname].layout = manager:GetLayout("main", uab.db.profile.layoutSettings[layoutname])
    end

    local layout = managers[layoutname].layout
    
    self:LayoutSettingsChanged()

	local config = layout:GetConfig()
	
	uab.options.args.lookandfeel.args.layout.args.configure.args = config:MenuOptionArgs()
--    self:GetCfgDlg():GetLayoutTab():SelectCurrentLayout()
end

function uab.classes.main.prototype:LayoutSettingsChanged()
    local layoutname = uab.db.profile.layout
    
    local managers = uab:GetManagers()
    local layout = managers[layoutname].layout

    for _, activator in pairs(self.activators) do
        activator:SetLayout(nil, layout)
    end
end

function uab.classes.main.prototype:ActionsChanged(activator)
    local layoutname = uab.db.profile.layout
    local managers = uab:GetManagers()
    local layout = managers[layoutname].layout

    if not activator then
        for _, activator in pairs(self.activators) do
            activator:SetLayout(nil, layout)
        end
    else
        activator:SetLayout(nil, layout)
    end
end

function uab.classes.main.prototype:ScaleChanged(groupname)
    for _, activator in pairs(self.activators) do
        activator:ScaleChanged(groupname)
    end
end

function uab.classes.main.prototype:PositionChanged(groupname)
    for _, activator in pairs(self.activators) do
        activator:PositionChanged(groupname)
    end
end

function uab.classes.main.prototype:DirectCastChanged()
    for _, activator in pairs(self.activators) do
        activator:DirectCastChanged()
    end
end

function uab.classes.main.prototype:StayOpenChanged()
    for _, activator in pairs(self.activators) do
        activator:StayOpenChanged()
    end
end

function uab.classes.main.prototype:FastChangeUnitChanged()
    for _, activator in pairs(self.activators) do
        activator:FastChangeUnitChanged()
    end
end

function uab.classes.main.prototype:ResetButtons()
    for _, activator in pairs(self.activators) do
        activator:ResetButtons()
    end
end

function uab.classes.main.prototype:FindNextAvailableActivation()
    --[[
    Find the first click combination that is not already used by
    some action set.  Search for the combination in the following order:

    The modifiers are tried in the following order for each button:
        
    shift
    ctrl
    alt
    shift-ctrl
    shift-alt
    ctrl-alt
    ctrl-shift-alt
    
    The buttons are tried in the following order:
    
    RightButton
    LeftButton
    MiddleButton
    Button4
    Button5
    
    If none of these match, the button alone is tried in the following
    order:
    
    MiddleButton
    Button4
    Button5
    RightButton
    LeftButton

    If all of these are taken, an error is generated, and no action set
    can be created.    
    --]]
    
    for idx, entry in pairs(uab.classes.ActionsTab.activations) do
        if not self:AnyActivatorWithActivation(entry) then
            return entry
        end
    end
end
            
function uab.classes.main.prototype:AnyActivatorWithActivation(entry)
    for name, activator in pairs(self.activators) do
        local settings = uab.db.profile.activators[name]
        
        if settings.button == entry.button and
            settings.prefixshift == entry.prefixshift and
            settings.prefixctrl == entry.prefixctrl and
            settings.prefixalt == entry.prefixalt then
            return true
        end
    end
end

