local print = function(msg) if msg then DEFAULT_CHAT_FRAME:AddMessage(msg) end end
--local pdebug = function(msg) if msg then table.insert(UABDebug, msg) end end
--local pdebug = function(msg) print(msg) end
local pdebug = function(msg) end

local UAB_GROW_UP = "u"
local UAB_GROW_DOWN = "d"
local uab.constants.grow.left = "l"
local uab.constants.grow.right = "r"

local uab.constants.placement.topleft = "tl"
local uab.constants.placement.top = "t"
local uab.constants.placement.topright = "tr"
local uab.constants.placement.bottomleft = "bl"
local uab.constants.placement.bottom = "b"
local uab.constants.placement.bottomright = "br"
local uab.constants.placement.left = "l"
local uab.constants.placement.right = "r"

--[[
-----------------------------------------------------------------------------------
Class: UnitActionBars.layouts.uabbar
Static Methods:
Instance Methods:
-----------------------------------------------------------------------------------
--]]
UnitActionBars.layouts.uabbar = UnitActionBars.aceOO.Class(UnitActionBars.layoutManager.ILayout, UnitActionBars.aceEvent)

function UnitActionBars.layouts.uabbar.GetLayoutName()
    return "UABBar"
end

function UnitActionBars.layouts.uabbar.GetDisplayName()
    return "Bar"
end

function UnitActionBars.layouts.uabbar.GetConfigTemplate()
    return "UAB_BarLayoutOptionsTemplate"
end
     
function UnitActionBars.layouts.uabbar.prototype:init(configFrame, sampleFrame, settings)
    UnitActionBars.layouts.uabbar.super.prototype.init(self)

    self:RegisterEvent("UnitActionBars_CreateGroupUnit", "CreateGroupUnit")

    self.configFrame = configFrame
    self.sampleFrame = sampleFrame
    configFrame.layout = self
    
    self.controllers = {}
    self.settings = UnitActionBars.layouts.uabbar.settings:new(configFrame, settings)
    self.sampleController = UnitActionBars.layouts.uabbar.controller:new(self.settings, "sample", 1, sampleFrame, 1)

    configFrame:SetScript("OnShow", function(self) self.layout:ShowSample() end)
    configFrame:SetScript("OnHide", function(self) self.layout:HideSample() end)
end

function UnitActionBars.layouts.uabbar.prototype:ShowSample()
    self.sampleController:EnableActionSet()
end

function UnitActionBars.layouts.uabbar.prototype:HideSample()
    self.sampleController:DisableActionSet()
end

function UnitActionBars.layouts.uabbar.prototype:CreateGroupUnit(name, id, header)
    self.controllers[name..id] = UnitActionBars.layouts.uabbar.controller:new(self.settings, name, id, header)
end

function UnitActionBars.layouts.uabbar.prototype:GetSettings()
    self.settings:GetValues()
end

function UnitActionBars.layouts.uabbar.prototype:DefaultSettings(bypassNotification)
    self.settings:Defaults(bypassNotification)
end

--[[
-----------------------------------------------------------------------------------
Class: UnitActionBars.layouts.uabbar.controller
Static Methods:
Instance Methods:
-----------------------------------------------------------------------------------
--]]
UnitActionBars.layouts.uabbar.controller = UnitActionBars.aceOO.Class(UnitActionBars.aceEvent)

function UnitActionBars.layouts.uabbar.controller.prototype:init(settings, name, id, header, sample)
    UnitActionBars.layouts.uabbar.controller.super.prototype.init(self)
    
    self.id = id
    self.settings = settings
    self.sample = sample
    self.header = header
    self.groupname = name
    self.reposition = 1
    self.revisual = 1

    local actionsetname
    
    if sample then
        actionsetname = header:GetName().."UABBarSample"
    
        self.header = CreateFrame("Frame", actionsetname.."Frame", header)
        self.header:Hide()
        self.header:SetHeight(header:GetHeight())
        self.header:SetWidth(header:GetWidth())
        self.header:SetPoint("CENTER")
    end
    
    self.actionset = UnitActionBars.actionsets.standard:GetActionSet(self.header, id, actionsetname)

    self:RegisterEvent("UnitActionBars_UABBar_SettingsChanged", "SettingsChanged")
    self:RegisterEvent("UnitActionBars_"..self.groupname.."_ChangeVisualActionEffect", "ChangeVisualEffect")
    self:RegisterEvent("UnitActionBars_UABBar_Enable", "EnableActionSet")
    self:RegisterEvent("UnitActionBars_UABBar_Disable", "DisableActionSet")
    
    if not sample then
        self:RegisterEvent("UnitActionBars_"..name.."_SetAction", "SetAction")
        self:RegisterEvent("UnitActionBars_"..name.."_RemoveAction", "RemoveAction")
    end

    self.actionset:SetInUse()
    if self.sample then
        self.header:Hide()
        for idx = 1, 24, 1 do
            self.actionset:SetAction(idx)
        end
    end
end

function UnitActionBars.layouts.uabbar.controller.prototype:SettingsChanged(settings)
    self.settings = settings
    self.reposition = 1
    self:UpdatePositions()
end

function UnitActionBars.layouts.uabbar.controller.prototype:ChangeVisualEffect(id, effects)
    self.revisual = 1
    self:TriggerEvent("UnitActionBars_"..self.actionset:ActionSetName().."_ChangeVisualActionEffect"..id, effects)
end

function UnitActionBars.layouts.uabbar.controller.prototype:EnableActionSet()
    if not self.enabled then
        self.enabled = 1
        self.actionset:SetController(self)
        self.actionset:SetInUse(1)
        self.reposition = 1
        self.revisual = 1
        self:UpdateAll()
        
        if self.sample then
            self.header:Show()
        end
    end
end

function UnitActionBars.layouts.uabbar.controller.prototype:DisableActionSet()
    if self.enabled then
        self.enabled = nil    
        self.actionset:SetInUse()
        if self.sample then
            self.header:Hide()
        end
    end
end

function UnitActionBars.layouts.uabbar.controller.prototype:SetAction(number, action)
    self.reposition = 1
    self.revisual = 1
    self.actionset:SetAction(number, action)
end

function UnitActionBars.layouts.uabbar.controller.prototype:RemoveAction(number)
    if self.actionset:ActionButton(number) then
        self.reposition = 1
        self.revisual = 1
        self.actionset:RemoveAction(number)
    end
end

function UnitActionBars.layouts.uabbar.controller.prototype:UpdateAll()
    self:UpdatePositions()
    self:UpdateVisualEffects()
end

function UnitActionBars.layouts.uabbar.controller.prototype:UpdatePositions()
    if self.reposition and self.actionset:IsVisible() then
        self.reposition = nil

        if self.sample then
            self:RepositionSampleCancel()
        end
        
        -- repositioning logic goes here
        self.actionset:ClearPositions()        

        local overflowModifier = { x = 0, y = 0 }
        local buttonRows = 24 / self.settings.growLimit
    
        if self.settings.overflowClockwise == "true" then
            if self.settings.growDirection == uab.constants.grow.up then
                overflowModifier.x = 1
                overflowModifier.y = 0
            elseif self.settings.growDirection == uab.constants.grow.down then
                overflowModifier.x = -1
                overflowModifier.y = 0
            elseif self.settings.growDirection == uab.constants.grow.right then
                overflowModifier.x = 0
                overflowModifier.y = -1
            else
                overflowModifier.x = 0
                overflowModifier.y = 1
            end
        else
            if self.settings.growDirection == uab.constants.grow.up then
                overflowModifier.x = -1
                overflowModifier.y = 0
            elseif self.settings.growDirection == uab.constants.grow.down then
                overflowModifier.x = 1
                overflowModifier.y = 0
            elseif self.settings.growDirection == uab.constants.grow.right then
                overflowModifier.x = 0
                overflowModifier.y = 1
            else
                overflowModifier.x = 0
                overflowModifier.y = -1
            end
        end

        local growModifier = { x = 0, y = 0 }
        local start = { x = 0, y = 0 }
        
        if self.settings.placement == uab.constants.placement.top then
            if self.settings.growDirection == uab.constants.grow.up then
                growModifier.x = 0
                growModifier.y = 1
            elseif self.settings.growDirection == uab.constants.grow.left then
                growModifier.x = -1
                growModifier.y = 0
            else    -- uab.constants.grow.right
                growModifier.x = 1
                growModifier.y = 0
            end
    
            start.x = 0
            start.y = 1
        elseif self.settings.placement == uab.constants.placement.right then
            if self.settings.growDirection == uab.constants.grow.right then
                growModifier.x = 1
                growModifier.y = 0
            elseif self.settings.growDirection == uab.constants.grow.up then
                growModifier.x = 0
                growModifier.y = 1
            else    -- uab.constants.grow.down
                growModifier.x = 0
                growModifier.y = -1
            end
    
            start.x = 1
            start.y = 0
        elseif self.settings.placement == uab.constants.placement.bottom then
            if self.settings.growDirection == uab.constants.grow.down then
                growModifier.x = 0
                growModifier.y = -1
            elseif self.settings.growDirection == uab.constants.grow.right then
                growModifier.x = 1
                growModifier.y = 0
            else    -- uab.constants.grow.left
                growModifier.x = -1
                growModifier.y = 0
            end
    
            start.x = 0
            start.y = -1
        elseif self.settings.placement == uab.constants.placement.left then
            if self.settings.growDirection == uab.constants.grow.left then
                growModifier.x = -1
                growModifier.y = 0
            elseif self.settings.growDirection == uab.constants.grow.up then
                growModifier.x = 0
                growModifier.y = 1
            else    -- uab.constants.grow.down
                growModifier.x = 0
                growModifier.y = -1
            end
    
            start.x = -1
            start.y = 0
        end

        local positions = {}
        local cancel = self.actionset:CancelButton()
        local button = self.actionset:ActionButton(1)
        
        positions[1] = { x = start.x * (UnitActionBars.actionbuttons.standard.buttonWidth + 1), y = start.y * (UnitActionBars.actionbuttons.standard.buttonHeight + 1) }

        if button then
            button:SetPoint("CENTER", cancel, "CENTER", positions[1].x, positions[1].y)
        end

        for pos = 2, self.settings.growLimit, 1 do
            positions[pos] = { x = 0, y = 0 }
            
            positions[pos].x = (start.x + (growModifier.x * (pos - 1))) * (UnitActionBars.actionbuttons.standard.buttonWidth + 1)
            positions[pos].y = (start.y + (growModifier.y * (pos - 1))) * (UnitActionBars.actionbuttons.standard.buttonHeight + 1)

            button = self.actionset:ActionButton(pos)

            if button then
                button:SetPoint("CENTER", cancel, "CENTER", positions[pos].x, positions[pos].y)
            end
        end

        for row = 2, buttonRows, 1 do
            self:PositionRow(positions, row, overflowModifier)
        end
    end
end

function UnitActionBars.layouts.uabbar.controller.prototype:RepositionSampleCancel()
    local cancelbtn = self.actionset:CancelButton()
    
    cancelbtn:ClearAllPoints()

    if self.settings.placement == uab.constants.placement.top then
        if self.settings.growDirection == uab.constants.grow.up then
            -- put the cancel button near the bottom of the parent frame, since
            --  we are putting the actions on top of the button, and growing
            --	them up
            cancelbtn:SetPoint("BOTTOM", self.header, 0, (UnitActionBars.layouts.uabbar.controller.buttonHeight))
        elseif self.settings.growDirection == uab.constants.grow.left then
            -- put the cancel button near the right side of the parent frame,
            --  since we are putting the actions on top of the button, and
            --  growing them to the left
            cancelbtn:SetPoint("RIGHT", self.header, -(UnitActionBars.actionbuttons.standard.buttonWidth), 0)
        else    -- uab.constants.grow.right
            -- put the cancel button near the left side of the parent frame,
            --  since we are putting the actions on top of the button, and
            --  growing them to the right
            cancelbtn:SetPoint("LEFT", self.header, (UnitActionBars.actionbuttons.standard.buttonWidth), 0)
        end
    elseif self.settings.placement == uab.constants.placement.right then
        if self.settings.growDirection == uab.constants.grow.right then
            -- put the cancel button near the left side of the parent frame,
            --  since we are putting the actions on the right side of the
            --  button, and growing them to the right
            cancelbtn:SetPoint("LEFT", self.header, (UnitActionBars.actionbuttons.standard.buttonWidth), 0)
        elseif self.settings.growDirection == uab.constants.grow.up then
            -- put the cancel button near the bottom of the parent frame, since
            --  we are putting the actions on the right side of the button, and
            --  growing them up
            cancelbtn:SetPoint("BOTTOM", self.header, 0, (UnitActionBars.actionbuttons.standard.buttonHeight))
        else    -- uab.constants.grow.down
            -- put the cancel button near the top of the parent frame,
            --  since we are putting the actions on the right side of the
            --  button, and growing them down
            cancelbtn:SetPoint("TOP", self.header, 0, -(UnitActionBars.actionbuttons.standard.buttonHeight))
        end
    elseif self.settings.placement == uab.constants.placement.bottom then
        if self.settings.growDirection == uab.constants.grow.down then
            -- put the cancel button near the top of the parent frame,
            --  since we are putting the actions under the button, and growing
            --  them down
            cancelbtn:SetPoint("TOP", self.header, 0, -(UnitActionBars.actionbuttons.standard.buttonHeight))
        elseif self.settings.growDirection == uab.constants.grow.right then
            -- put the cancel button near the left side of the parent frame,
            --  since we are putting the actions under the button, and growing
            --  them to the right
            cancelbtn:SetPoint("LEFT", self.header, (UnitActionBars.actionbuttons.standard.buttonWidth), 0)
        else    -- uab.constants.grow.left
            -- put the cancel button near the right side of the parent frame,
            --  since we are putting the actions under the button, and growing
            --  them to the left
            cancelbtn:SetPoint("RIGHT", self.header, -(UnitActionBars.actionbuttons.standard.buttonWidth), 0)
        end
    elseif self.settings.placement == uab.constants.placement.left then
        if self.settings.growDirection == uab.constants.grow.left then
            -- put the cancel button near the right side of the parent frame,
            --  since we are putting the actions on the left side of the button,
            --  and growing them to the left
            cancelbtn:SetPoint("RIGHT", self.header, -(UnitActionBars.actionbuttons.standard.buttonWidth), 0)
        elseif self.settings.growDirection == uab.constants.grow.up then
            -- put the cancel button near the bottom of the parent frame, since
            --  we are putting the actions on the left side of the button, and
            --  growing them up
            cancelbtn:SetPoint("BOTTOM", self.header, 0, (UnitActionBars.actionbuttons.standard.buttonHeight))
        else    -- uab.constants.grow.down
            -- put the cancel button near the top of the parent frame,
            --  since we are putting the actions on the left side of the
            --  button, and growing them down
            cancelbtn:SetPoint("TOP", self.header, 0, -(UnitActionBars.actionbuttons.standard.buttonHeight))
        end
    end
end

function UnitActionBars.layouts.uabbar.controller.prototype:PositionRow(positions, row, placementMod)
    local ofs = (row - 1) * self.settings.growLimit
    local relofs = ofs - self.settings.growLimit

    for pos = 1, self.settings.growLimit, 1 do
        local btnpos = pos + ofs
        local btn = self.actionset:ActionButton(btnpos)
        local relbtnpos = pos + relofs
        local relpos = positions[relbtnpos]
        
        positions[btnpos] = { x = 0, y = 0 }

        positions[btnpos].x = relpos.x + (placementMod.x * (UnitActionBars.actionbuttons.standard.buttonWidth + 1))
        positions[btnpos].y = relpos.y + (placementMod.y * (UnitActionBars.actionbuttons.standard.buttonHeight + 1))
                    
        if btn then
            btn:SetPoint("CENTER", self.actionset:CancelButton(), "CENTER", positions[btnpos].x, positions[btnpos].y)
        end
    end
end

function UnitActionBars.layouts.uabbar.controller.prototype:UpdateVisualEffects()
    if self.revisual and self.actionset:IsVisible() then
        self.revisual = nil
    
        -- apply all the visual properties based on user selections (low alpha, highlight border, blink)
        -- for each button control in the actionset

        --self.actionset:UpdateVisualEffects()        
    end
end

--[[
-----------------------------------------------------------------------------------
Class: UnitActionBars.layouts.uabbar.settings
Static Methods:
Instance Methods:
-----------------------------------------------------------------------------------
--]]
UnitActionBars.layouts.uabbar.settings = UnitActionBars.aceOO.Class(UnitActionBars.aceEvent)

function UnitActionBars.layouts.uabbar.settings.PlacementDisplayValue(value)
    if value == uab.constants.placement.top then
        return "Top"
    elseif value == uab.constants.placement.bottom then
        return "Bottom"
    elseif value == uab.constants.placement.left then
        return "Left"
    else
        return "Right"
    end
end

function UnitActionBars.layouts.uabbar.settings.GrowDirectionDisplayValue(value)
    if value == uab.constants.grow.up then
        return "Up"
    elseif value == uab.constants.grow.down then
        return "Down"
    elseif value == uab.constants.grow.left then
        return "Left"
    else
        return "Right"
    end
end

function UnitActionBars.layouts.uabbar.settings.GrowLimitDisplayValue(value)
    return tostring(value)
end

function UnitActionBars.layouts.uabbar.settings.prototype:init(configFrame, settings)
    UnitActionBars.layouts.uabbar.settings.super.prototype.init(self)
    
    self.frame = configFrame

    self.overflowClockwise = (settings and settings.overflowClockwise) or "true"
    self.placement = (settings and settings.placement) or uab.constants.placement.right
    self.growDirection = (settings and settings.growDirection) or uab.constants.grow.right
    self.growLimit = (settings and settings.growLimit) or 4

	self.placementFrame = getglobal(self.frame:GetName().."Placement")
	self.placementFrame.settings = self

    UnitActionBars.layouts.uabbar.settings.InsideInit = 1
    UnitActionBars.layouts.uabbar.settings.InitDropdown = self
    
    UIDropDownMenu_Initialize(self.placementFrame,
        function()
            if UnitActionBars.layouts.uabbar.settings.InsideInit then
                UnitActionBars.layouts.uabbar.settings.InitDropdown:PlacementListInitialize()
            else
                this:GetParent().settings:PlacementListInitialize()
            end
        end)
	UIDropDownMenu_SetWidth(110, self.placementFrame)
	UIDropDownMenu_SetButtonWidth(24, self.placementFrame)
	UIDropDownMenu_JustifyText("LEFT", self.placementFrame)

	self.growDirectionFrame = getglobal(self.frame:GetName().."GrowDirection")
	self.growDirectionFrame.settings = self

    UIDropDownMenu_Initialize(self.growDirectionFrame,
        function()
            if UnitActionBars.layouts.uabbar.settings.InsideInit then
                UnitActionBars.layouts.uabbar.settings.InitDropdown:GrowDirectionListInitialize()
            else
                this:GetParent().settings:GrowDirectionListInitialize()
            end
        end)
	UIDropDownMenu_SetWidth(110, self.growDirectionFrame)
	UIDropDownMenu_SetButtonWidth(24, self.growDirectionFrame)
	UIDropDownMenu_JustifyText("LEFT", self.growDirectionFrame)

	self.growLimitFrame = getglobal(self.frame:GetName().."GrowLimit")
	self.growLimitFrame.settings = self

    UIDropDownMenu_Initialize(self.growLimitFrame,
        function()
            if UnitActionBars.layouts.uabbar.settings.InsideInit then
                UnitActionBars.layouts.uabbar.settings.InitDropdown:GrowLimitListInitialize()
            else
                this:GetParent().settings:GrowLimitListInitialize()
            end
        end)
	UIDropDownMenu_SetWidth(110, self.growLimitFrame)
	UIDropDownMenu_SetButtonWidth(24, self.growLimitFrame)
	UIDropDownMenu_JustifyText("LEFT", self.growLimitFrame)

    UnitActionBars.layouts.uabbar.settings.InsideInit = nil
    UnitActionBars.layouts.uabbar.settings.InitDropdown = nil

	self.overflowDirectionFrame = getglobal(self.frame:GetName().."OverflowDirection")
    self.overflowDirectionFrame.settings = self
    
	getglobal(self.overflowDirectionFrame:GetName().."Text"):SetText("Overflow buttons clockwise")
	self.overflowDirectionFrame:SetScript("OnClick", function(self, button, down) self.settings:OverflowDirectionButtonClicked(self) end)

	self:SetOverflowClockwise(self.overflowClockwise)
	self:SetPlacement(self.placement)
	self:SetGrowDirection(self.growDirection)
	self:SetGrowLimit(self.growLimit)

    self:RegisterEvent("UnitActionBars_ProfileChanged", "ProfileChanged")    
    self:TriggerEvent("UnitActionBars_UABBar_SettingsChanged", self)
end

function UnitActionBars.layouts.uabbar.settings.prototype:ProfileChanged()
    local settings = UnitActionBars.db.profile.layoutSettings.UABBar
    
    self.overflowClockwise = (settings and settings.overflowClockwise) or "true"
    self.placement = (settings and settings.placement) or uab.constants.placement.right
    self.growDirection = (settings and settings.growDirection) or uab.constants.grow.right
    self.growLimit = (settings and settings.growLimit) or 4
    
	self:SetOverflowClockwise(self.overflowClockwise)
	self:SetPlacement(self.placement)
	self:SetGrowDirection(self.growDirection)
	self:SetGrowLimit(self.growLimit)
end

function UnitActionBars.layouts.uabbar.settings.prototype:Defaults(bypassNotification)
    self.overflowClockwise = "true"
    self.placement = uab.constants.placement.right
    self.growDirection = uab.constants.grow.right
    self.growLimit = 4
    
    UnitActionBars.db.profile.layoutSettings.UABBar = self:GetValues()
    
    if not bypassNotification then
        self:TriggerEvent("UnitActionBars_UABBar_SettingsChanged", self)
    end
end

function UnitActionBars.layouts.uabbar.settings.prototype:PlacementListInitialize()
    local index = 1
    local selected = nil

    local options = {}

    table.insert(options, {	text = "Top",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectPlacement", arg1 = uab.constants.placement.top }})
    table.insert(options, {	text = "Right",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectPlacement", arg1 = uab.constants.placement.right }})
    table.insert(options, {	text = "Bottom",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectPlacement", arg1 = uab.constants.placement.bottom }})
    table.insert(options, {	text = "Left",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectPlacement", arg1 = uab.constants.placement.left }})

    for index, data in pairs(options) do
        UIDropDownMenu_AddButton(data)
        if self.placement == data.arg2.arg1 then
            selected = index
        end
    end

    UIDropDownMenu_SetSelectedID(self.placementFrame, selected)
end

function UnitActionBars.layouts.uabbar.settings.prototype:SelectPlacement(selector, selection)
    UIDropDownMenu_SetSelectedID(self.placementFrame, selector:GetID())
    self:SetPlacement(selection)
end

function UnitActionBars.layouts.uabbar.settings.prototype:SetPlacement(value)
    local notify = (value ~= self.placement)
                
    self.placement = value
        
    UIDropDownMenu_SetText(UnitActionBars.layouts.uabbar.settings.PlacementDisplayValue(value), self.placementFrame)
    self:SynchronizeGrowPattern()
    
    UnitActionBars.db.profile.layoutSettings.UABBar = self:GetValues()
    
    if notify then
        self:TriggerEvent("UnitActionBars_UABBar_SettingsChanged", self)
    end        
end

function UnitActionBars.layouts.uabbar.settings.prototype:GrowDirectionListInitialize()
    local index = 1
    local selected = nil

    local options = {}

    table.insert(options, {	text = "Up",
                            func = UAB_DropDownSelect,
                            arg1 =  self,
                            arg2 = { func = "SelectGrowDirection", arg1 = uab.constants.grow.up }})
    table.insert(options, {	text = "Right",
                            func = UAB_DropDownSelect,
                            arg1 =  self,
                            arg2 = { func = "SelectGrowDirection", arg1 = uab.constants.grow.right }})
    table.insert(options, {	text = "Down",
                            func = UAB_DropDownSelect,
                            arg1 =  self,
                            arg2 = { func = "SelectGrowDirection", arg1 = uab.constants.grow.down }})
    table.insert(options, {	text = "Left",
                            func = UAB_DropDownSelect,
                            arg1 =  self,
                            arg2 = { func = "SelectGrowDirection", arg1 = uab.constants.grow.left }})

    for index, data in pairs(options) do
        UIDropDownMenu_AddButton(data)
        if self.growDirection == data.arg2.arg1 then
            selected = index
        end
    end

    UIDropDownMenu_SetSelectedID(self.growDirectionFrame, selected)
end

function UnitActionBars.layouts.uabbar.settings.prototype:SelectGrowDirection(selector, selection)
    UIDropDownMenu_SetSelectedID(self.growDirectionFrame, selector:GetID())
    self:SetGrowDirection(selection)
end

function UnitActionBars.layouts.uabbar.settings.prototype:SetGrowDirection(value)
    local notify = (value ~= self.growDirection)
    
    self.growDirection = value
    
    UIDropDownMenu_SetText(UnitActionBars.layouts.uabbar.settings.GrowDirectionDisplayValue(value), self.growDirectionFrame)

    self:SynchronizeGrowPattern()
    
    UnitActionBars.db.profile.layoutSettings.UABBar = self:GetValues()
    
    if notify then
        self:TriggerEvent("UnitActionBars_UABBar_SettingsChanged", self)
    end        
end

function UnitActionBars.layouts.uabbar.settings.prototype:GrowLimitListInitialize()
    local index = 1
    local selected = nil

    local options = {}

    table.insert(options, {	text = "2",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectGrowLimit", arg1 = "2" }})
    table.insert(options, {	text = "3",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectGrowLimit", arg1 = "3" }})
    table.insert(options, {	text = "4",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectGrowLimit", arg1 = "4" }})
    table.insert(options, {	text = "6",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectGrowLimit", arg1 = "6" }})
    table.insert(options, {	text = "8",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectGrowLimit", arg1 = "8" }})
    table.insert(options, {	text = "12",
                            func = UAB_DropDownSelect,
                            arg1 = self,
                            arg2 = { func = "SelectGrowLimit", arg1 = "12" }})

    for index, data in pairs(options) do
        UIDropDownMenu_AddButton(data)
        if self.growLimit == tonumber(data.arg2.arg1) then
            selected = index
        end
    end

    UIDropDownMenu_SetSelectedID(self.growLimitFrame, selected)
end

function UnitActionBars.layouts.uabbar.settings.prototype:SelectGrowLimit(selector, selection)
    UIDropDownMenu_SetSelectedID(self.growLimitFrame, selector:GetID())
    self:SetGrowLimit(selection)
end

function UnitActionBars.layouts.uabbar.settings.prototype:SetGrowLimit(value)
    local notify = (tonumber(value) ~= self.growLimit)
    
    self.growLimit = tonumber(value)
    UIDropDownMenu_SetText(UnitActionBars.layouts.uabbar.settings.GrowLimitDisplayValue(value), self.growLimitFrame)
    
    UnitActionBars.db.profile.layoutSettings.UABBar = self:GetValues()
    
    if notify then
        self:TriggerEvent("UnitActionBars_UABBar_SettingsChanged", self)
    end        
end

function UnitActionBars.layouts.uabbar.settings.prototype:OverflowDirectionButtonClicked(frame)
    self:SetOverflowClockwise(frame:GetChecked())
end

function UnitActionBars.layouts.uabbar.settings.prototype:SetOverflowClockwise(value)
    local newval
    
    if value and (tonumber(value) == 1 or value == "true") then
        newval = "true"
    else
        newval = "false"
    end
    
    local notify = (newval ~= self.overflowClockwise)
        
    self.overflowClockwise = newval
    self.overflowDirectionFrame:SetChecked(self.overflowClockwise == "true")
    
    UnitActionBars.db.profile.layoutSettings.UABBar = self:GetValues()
    
    if notify then
        self:TriggerEvent("UnitActionBars_UABBar_SettingsChanged", self)
    end        
end

function UnitActionBars.layouts.uabbar.settings.prototype:SynchronizeGrowPattern()
    self.overflowDirectionFrame:Enable()
    if self.placement and self.growDirection then
        if self.placement == uab.constants.placement.top then
            if self.growDirection == uab.constants.grow.down then
                self:SetGrowDirection(uab.constants.grow.up)
            elseif self.growDirection == uab.constants.grow.right then
                self:SetOverflowClockwise("false")
                self.overflowDirectionFrame:Disable()
            elseif self.growDirection == uab.constants.grow.left then
                self:SetOverflowClockwise("true")
                self.overflowDirectionFrame:Disable()
            end
        elseif self.placement == uab.constants.placement.bottom then
            if self.growDirection == uab.constants.grow.up then
                self:SetGrowDirection(uab.constants.grow.down)
            elseif self.growDirection == uab.constants.grow.right then
                self:SetOverflowClockwise("true")
                self.overflowDirectionFrame:Disable()
            elseif self.growDirection == uab.constants.grow.left then
                self:SetOverflowClockwise("false")
                self.overflowDirectionFrame:Disable()
            end
        elseif self.placement == uab.constants.placement.left then
            if self.growDirection == uab.constants.grow.right then
                self:SetGrowDirection(uab.constants.grow.left)
            elseif self.growDirection == uab.constants.grow.up then
                self:SetOverflowClockwise("false")
                self.overflowDirectionFrame:Disable()
            elseif self.growDirection == uab.constants.grow.down then
                self:SetOverflowClockwise("true")
                self.overflowDirectionFrame:Disable()
            end
        elseif self.placement == uab.constants.placement.right then
            if self.growDirection == uab.constants.grow.left then
                self:SetGrowDirection(uab.constants.grow.right)
            elseif self.growDirection == uab.constants.grow.up then
                self:SetOverflowClockwise("true")
                self.overflowDirectionFrame:Disable()
            elseif self.growDirection == uab.constants.grow.down then
                self:SetOverflowClockwise("false")
                self.overflowDirectionFrame:Disable()
            end
        end
    end
end

function UnitActionBars.layouts.uabbar.settings.prototype:GetValues()
    local values = {}
    
    values.overflowClockwise = self.overflowClockwise
    values.placement = self.placement
    values.growDirection = self.growDirection
    values.growLimit = self.growLimit
    
    return values
end

UnitActionBars.layoutManager:Register( UnitActionBars.layouts.uabbar )
